import inspect
from abc import ABC, abstractmethod
from typing import Optional

from ._globals import LoadType


class Context(ABC):
    def __init__(self, load_type: LoadType):
        self.__load_type = load_type

    @property
    def load_type(self):
        return self.__load_type

    @abstractmethod
    def predicate(self):
        raise Exception("'predicate' method is not defined.")

    @abstractmethod
    def draw_comparison(self, file: str):
        raise Exception("'draw_comparison' method is not defined.")


class _ClassContext(Context):
    def predicate(self):
        return inspect.isclass

    def draw_comparison(self, file: str):
        return "".join([s.capitalize() for s in file.split("_")])


class _FunctionContext(Context):
    def predicate(self):
        return inspect.isfunction

    def draw_comparison(self, file: str):
        return file.lower()


class ContextFactory:
    __class_context: Optional[_ClassContext] = None
    __function_context: Optional[_FunctionContext] = None

    @classmethod
    def get(cls, load_type: LoadType) -> Context:
        if load_type == LoadType.clazz:
            if cls.__class_context is None:
                cls.__class_context = _ClassContext(load_type)
            return cls.__class_context
        if cls.__function_context is None:
            cls.__function_context = _FunctionContext(load_type)
        return cls.__function_context



import inspect
import warnings
from dataclasses import dataclass
from os import path as os_path
from typing import Callable, ClassVar, Iterable, List, Optional, Tuple, Type

from ._context import Context, ContextFactory
from ._globals import Class_Or_Func, LoadType
from ._import import ImportableFactory, ImportOption

__all__ = ("ModuleLoader", "ModuleLoaderSetting")


class __Private:
    """Private namespace.
    If you do not want to expose, define here as much as possible.
    """

    THIS_FILE = os_path.basename(__file__)
    DEFAULT_EXCLUDES = ("__init__.py", THIS_FILE, "__pycache__")

    def __new__(cls, *args, **kwargs):
        raise Exception(f"{cls.__name__} can't be initialized.")

    @classmethod
    def detect_call_path(cls):
        this_file = cls.THIS_FILE
        stack = inspect.stack()
        for path in stack:
            path_name = path.filename
            if this_file == os_path.basename(path_name):
                continue
            return path_name

    @classmethod
    def init_base_url(cls, base_path: Optional[str] = None):
        if base_path is None:
            return cls.init_base_url(os_path.dirname(cls.detect_call_path()))
        if base_path == "/":
            return base_path
        if base_path == "":
            return "/"
        if base_path.endswith("/"):
            return base_path[:-1]
        return base_path


def _access_private():
    return __Private


@dataclass(frozen=True)
class ModuleLoaderSetting:
    base_path: Optional[str] = None
    strict: bool = False
    singleton: bool = False


class ModuleLoader:
    _setting: ClassVar[ModuleLoaderSetting] = ModuleLoaderSetting()
    _instance: Optional["ModuleLoader"] = None

    @classmethod
    def get_setting(cls) -> ModuleLoaderSetting:
        return cls._setting

    @classmethod
    def set_setting(
        cls,
        base_path: Optional[str] = None,
        strict: bool = False,
        singleton: bool = False,
    ) -> None:
        cls._setting = ModuleLoaderSetting(base_path, strict, singleton)

    def __new__(cls, *args, **kwargs):
        if cls._setting.singleton is False:
            return super(ModuleLoader, cls).__new__(cls)
        if cls._instance is None:
            cls._instance = super(ModuleLoader, cls).__new__(cls)
        return cls._instance

    def __init__(self, base_path: Optional[str] = None, strict: Optional[bool] = None):
        """initialize
        :param base_path: Base path for import.
            Defaults to the path where this object was initialized.
        :param strict: If strict is True,
            ModuleLoader strictly try to load a class or function object
            per a Python module on a basis of its name.
        """
        setting = ModuleLoader._setting
        if setting.singleton is True and hasattr(self, "_ModuleLoader__base_path"):
            return
        global_base_path, global_strict = setting.base_path, setting.strict
        self.__base_path: str = (
            _access_private().init_base_url(base_path)
            if global_base_path is None
            else global_base_path
        )
        self.__strict: bool = global_strict if strict is None else strict

    @property
    def base_path(self) -> str:
        return self.__base_path

    @property
    def strict(self) -> bool:
        return self.__strict

    def load_class(self, file_name: str) -> Type:
        """Import Python module and return class.
        :param file_name: Python file name (Module name).
            You can input relative path like '../example' based on 'base_path'.
        :return: class object defined in the Python file (Module) according to rules.
        """
        return self.__load_resource(file_name, ContextFactory.get(LoadType.clazz))

    def load_function(self, file_name: str) -> Callable:
        """Import Python module and return function.
        :param file_name: Python file name (module name).
            You can input relative path like '../example' based on 'base_path'.
        :return: function object defined in the Python file (Module) according to rules.
        """
        return self.__load_resource(file_name, ContextFactory.get(LoadType.func))

    def load_classes(
        self,
        src: Optional[str] = None,  # Temporary Optional because of pkg_name
        excludes: Iterable[str] = (),
        recursive: bool = False,
        *args,
        **kwargs,
    ) -> Tuple[Type, ...]:
        """Import Python package and return classes.
        :param src: Python package or module name.
            You can input relative path like '../example' based on 'base_path'.
        :param excludes: Python file names you don't want to import in the package.
        :param recursive: If True, import Python package recursively.
        :return: class objects defined in the Python package according to rules.
        """
        pkg_name = kwargs.get("pkg_name")
        if kwargs.get("pkg_name") is not None:
            warnings.warn(
                "'pkg_name' is deprecated. Please use 'src' parameter.", FutureWarning
            )
            src = pkg_name
        if src is None:
            raise TypeError("'src' parameter is required.")
        return self.__load_resources(
            src,
            excludes=excludes,
            recursive=recursive,
            context=ContextFactory.get(LoadType.clazz),
        )

    def load_functions(
        self,
        src: Optional[str] = None,  # Temporary Optional because of pkg_name
        excludes: Iterable[str] = (),
        recursive: bool = False,
        *args,
        **kwargs,
    ) -> Tuple[Callable, ...]:
        """Import Python package and return functions.
        :param src: Python package or module name.
            You can input relative path like '../example' based on 'base_path'.
        :param excludes: Python file names you don't want to import in the package.
        :param recursive: If True, import Python package recursively.
        :return: function objects defined in the Python package according to rules.
        """
        pkg_name = kwargs.get("pkg_name")
        if kwargs.get("pkg_name") is not None:
            warnings.warn(
                "'pkg_name' is deprecated. Please use 'src' parameter.", FutureWarning
            )
            src = pkg_name
        if src is None:
            raise TypeError("'src' parameter is required.")
        return self.__load_resources(
            src,
            excludes=excludes,
            recursive=recursive,
            context=ContextFactory.get(LoadType.func),
        )

    def __path_fix(self, name: str) -> str:
        if not name or name == "." or name == "/" or name == "./":
            return self.__base_path
        if name.startswith("/"):
            result_path = self.__base_path + name
            # example: /foo/bar/
            if name.endswith("/"):
                return result_path[:-1]
            # example: /foo/bar
            return result_path
        if name.startswith("."):
            if name[1] != ".":
                if name[1] == "/":
                    result_path = self.__base_path + name[1:]
                    # example: ./foo/
                    if name.endswith("/"):
                        return result_path[:-1]
                    # example: ./foo
                    return result_path
                # example: .foo.bar
                return self.__base_path + "/".join(name.split("."))
            level = 0
            path = None
            for i in range(len(name)):
                if i == 0:
                    continue
                if name[i] != ".":
                    break
                path = name[i + 1 :]
                level += 1
            if path is not None:
                base_path_arr = self.__base_path.split("/")
                result_base_path = "/".join(
                    base_path_arr[0 : len(base_path_arr) - level]
                )
                if path.startswith("/"):
                    if path.endswith("/"):
                        # example: ../foo/
                        return result_base_path + path[:-1]
                    # example: ../foo
                    return result_base_path + path
                # example: ..foo.bar
                path = "/".join(path.split("."))
                return result_base_path + "/" + path
        # example: foo.bar
        path = "/".join(name.split("."))
        return self.__base_path + "/" + path

    def __load_resource(self, file_name: str, context: Context) -> Class_Or_Func:
        fix_path = self.__path_fix(file_name)
        importable = ImportableFactory.get(fix_path, context)
        return importable.import_resources()[0]

    def __load_resources(
        self,
        src: str,
        context: Context,
        excludes: Iterable[str] = (),
        recursive: bool = False,
    ) -> Tuple[Class_Or_Func, ...]:
        target_dir = self.__path_fix(src)
        private = _access_private()
        exclude_files = list(private.DEFAULT_EXCLUDES)
        exclude_files.append(os_path.basename(private.detect_call_path()))
        if excludes:
            if not iter(excludes):
                raise TypeError("excludes variable must be iterable.")
            for exclude in excludes:
                if not isinstance(exclude, str):
                    raise TypeError("The contents of the excludes must all be strings")
                exclude_files.append(exclude)
        import_option = ImportOption(recursive, exclude_files, self.__strict)
        importable = ImportableFactory.get(target_dir, context, import_option)
        mods: List[Class_Or_Func] = importable.import_resources()
        has_order_mods = [
            mod
            for mod in mods
            if hasattr(mod, "_load_order") and getattr(mod, "_load_order")
        ]
        if not has_order_mods:
            return tuple(mods)
        no_has_order_mods = [
            mod
            for mod in mods
            if not hasattr(mod, "_load_order") or not getattr(mod, "_load_order")
        ]
        if not no_has_order_mods:
            return tuple(
                sorted(has_order_mods, key=lambda m: getattr(m, "_load_order"))
            )
        ordered_mods = (
            sorted(has_order_mods, key=lambda m: getattr(m, "_load_order"))
            + no_has_order_mods
        )
        return tuple(ordered_mods)



import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib import import_module
from os import listdir
from os import path as os_path
from sys import path as sys_path
from typing import Iterable, List

from ._context import Context
from ._globals import Class_Or_Func
from .exception import LoaderStrictModeError


def _is_module(name: str) -> bool:
    return name.endswith(".py")


def _exclude_py(path: str) -> str:
    if _is_module(path):
        return path.replace(".py", "")
    return path


def _exclude_ex(file_name: str) -> str:
    return os_path.splitext(file_name)[0]


_DECORATOR_ATTR = "_load_flg"


@dataclass(frozen=True)
class ImportOption:
    recursive: bool = False
    excludes: Iterable[str] = ()
    is_strict: bool = False


default_option = ImportOption()


class Importable(ABC):
    def __init__(
        self, path: str, context: Context, option: ImportOption = default_option
    ):
        self._path = _exclude_py(path)
        self._context = context
        self._option = option
        children = self._load_children()
        fix_excludes = [_exclude_py(e) for e in option.excludes]
        self._children: List[Importable] = [
            child for child in children if child.get_base_name() not in fix_excludes
        ]

    @abstractmethod
    def import_resources(self) -> List[Class_Or_Func]:
        raise Exception("'import_resources' method is not defined.")

    def get_base_name(self) -> str:
        return _exclude_ex(os_path.basename(self._path))

    def _load_children(self) -> List["Importable"]:
        return []


class _Module(Importable):
    def import_resources(self) -> List[Class_Or_Func]:
        file = self.get_base_name()
        context = self._context
        is_strict = self._option.is_strict
        load_type_name = context.load_type.value
        module = import_module(file)
        target_load_name = context.draw_comparison(file)
        is_found = False
        error = None
        mods = []
        members = inspect.getmembers(module, context.predicate())
        for mod_name, mod in members:
            is_name_match = target_load_name == mod_name
            if hasattr(mod, _DECORATOR_ATTR):
                if not getattr(mod, _DECORATOR_ATTR):
                    continue
                if is_found:
                    # High priority error
                    error = LoaderStrictModeError(
                        f"Loader can only load a "
                        f"'{target_load_name}' {load_type_name} in {file} module."
                        f"\nPlease check '{mod_name}' in {file} module."
                    )
                    break
                if is_strict and not is_name_match:
                    error = LoaderStrictModeError(
                        f"Loader can't load '{mod_name}' in {file} module."
                        f"\nPlease rename '{target_load_name}' {load_type_name}."
                    )
                    continue
                mods.append(mod)
                if is_strict:
                    if error:
                        # High priority error
                        error = LoaderStrictModeError(
                            f"Loader can only load a "
                            f"'{target_load_name}' {load_type_name} "
                            f"in {file} module."
                        )
                        break
                    is_found = True
                continue
            if not is_name_match:
                continue
            mods.append(mod)
            if is_strict:
                if error:
                    # High priority error
                    error = LoaderStrictModeError(
                        f"Loader can only load a "
                        f"'{target_load_name}' {load_type_name} in {file} module."
                    )
                    break
                is_found = True
        if error is not None:
            raise error
        return mods


class _Package(Importable):
    def import_resources(self) -> List[Class_Or_Func]:
        return [
            resource
            for child in self._children
            for resource in child.import_resources()
        ]

    def _load_children(self) -> List[Importable]:
        path = self._path
        option = self._option
        children = []
        for file_or_dir in listdir(path):
            if file_or_dir in option.excludes:
                continue
            is_module = _is_module(file_or_dir)
            is_pkg = os_path.isdir(f"{path}/{file_or_dir}")
            if not is_module and not is_pkg:
                continue
            if is_pkg and not option.recursive:
                continue
            fixed_file_or_dir = _exclude_ex(file_or_dir) if is_module else file_or_dir
            children.append(
                ImportableFactory.get(
                    f"{path}/{fixed_file_or_dir}", self._context, option
                )
            )
        return children


class ImportableFactory:
    @staticmethod
    def get(path: str, *args, **kwargs) -> Importable:
        is_dir = os_path.isdir(path)
        fixed_path = path if is_dir else "/".join(path.split("/")[:-1])
        if fixed_path not in sys_path:
            sys_path.append(fixed_path)
        if is_dir:
            return _Package(path, *args, **kwargs)
        return _Module(path, *args, **kwargs)



from enum import Enum
from typing import Any, Callable, Type, TypeVar


class LoadType(Enum):
    func = "function"
    clazz = "class"


Class_Or_Func = TypeVar("Class_Or_Func", Type[Any], Callable)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autoload']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'autoload-module',
    'version': '1.7.0',
    'description': 'Python Autoload Module',
    'long_description': '# autoload_module\n[![PyPI version](https://badge.fury.io/py/autoload-module.svg)](https://badge.fury.io/py/autoload-module)\n[![Test](https://github.com/hiroki0525/autoload_module/actions/workflows/test.yml/badge.svg)](https://github.com/hiroki0525/autoload_module/actions/workflows/test.yml)\n<img src="https://img.shields.io/badge/license-MIT-blue.svg?style=flat" alt="MIT License image">\n\nThis library will give you comfortable Python metaprogramming.  \nThe following is a plain example.\n\n- Directory\n```\nproject/\n ├ example.py\n └ validator/\n   ├ validator_a.py\n   ├ validator_b.py\n   └ validator_c.py\n```\n- example.py\n```python\nfrom autoload import ModuleLoader\n\ninput = "foo bar baz"\nloader = ModuleLoader()\n\n# Automatically import modules and return class objects\nvalidator_classes = loader.load_classes("validator")\ntry:\n    # initialize and execute method\n    [clazz().validate(input) for clazz in validator_classes]\nexcept:\n    print("input is invalid!!")\n```\n## Install\n```\npip install autoload-module\n```\n## Usage\n### Constructor\n```\nModuleLoader(\n  base_path: Optional[str] = None,\n  strict: bool = False\n)\n```\nThe ModuleLoader can be generated with no parameters.\nIn that case, the instance has the absolute path where\nit was initialized.  \n- Directory\n```\n/usr/local/src/project/\n  ├ example.py\n  └ validator/\n    ├ validator_a.py\n    ├ validator_b.py\n    └ validator_c.py\n```\n- example.py\n```python\nfrom autoload import ModuleLoader\n\n# The instance has \'/usr/local/src/project/\'\nloader = ModuleLoader()\n\n# load modules in the directory; \'/usr/local/src/project/validator/\'\nvalidator_classes = loader.load_classes("validator")\n```\nIf you want to change the base path, you must generate the ModuleLoader with an absolute path parameter.\n```python\nloader = ModuleLoader(\'/user/local/src/custom\')\n```\n\nAbout strict parameter, please see [here](#NOTE) .\n\nYou can also create global setting and initialize singleton object.\n```python\nfrom autoload import ModuleLoader\nimport os\n\n# global setting\nModuleLoader.set_setting(base_path=os.getcwd(), strict=True)\n\nloader_a = ModuleLoader()\nloader_b = ModuleLoader()\n\nprint(loader_a.base_path)\n# -> /Users/user1/abc\nprint(loader_b.base_path)\n# -> /Users/user1/abc\n\n# singleton setting\nModuleLoader.set_setting(singleton=True)\n\nloader_c = ModuleLoader()\nloader_d = ModuleLoader()\nloader_e = ModuleLoader(\'/test\')\n\nassert loader_c is loader_d # OK\nassert loader_c is loader_e # OK\n\n# The base_path is \'/Users/user1/abc\'\nassert loader_c.base_path is loader_e.base_path # OK\n```\n\n### Methods\n#### load_classes\n```\nload_classes(\n    src: str,\n    excludes: Iterable[str] = (),\n    recursive: bool = False,\n) -> Tuple[Type]:\n```\nThis method read the Python package or module and return the tuple of class objects.\n\n**NOTE**\n\nFrom version 1.5.0, `pkg_name` parameter is duplicated.\nIt will be deleted soon. \n`load_functions` is also same.\n\n- Directory\n```\npkg/\n ├ example.py\n ├ __init__.py\n ├ config.yaml\n └ main/\n     ├ validator_a.py\n     ├ validator_b.py\n     ├ validator_c.py\n     └ sub/\n        ├ validator_d.py\n        └ validator_e.py\n```\n- validator_a.py\n```python\nclass ValidatorA:\n    def validate(self):\n        print("validateA!!")\n```\n- example.py\n```python\nloader = ModuleLoader()\n\n# Automatically read modules without \'__init__.py\', not py file, and this file.\n# return the tuple of ValidateA, ValidatorB, and ValidatorC class objects\nvalidator_classes = loader.load_classes("main")\n\n# initialize and execute method\n[clazz().validate() for clazz in validator_classes]\n# -> validateA!!\n# -> validateB!!\n# -> validateC!!\n```\nYou can also load only specific modules using `excludes` variable or `load_config` decorator as below.\n```python\n# Pattern1: \'excludes\'\n# \'excludes\' is a iterable object like tuple, list.\n# You must specify module names in \'excludes\'.\nvalidator_classes = loader.load_classes("main", ["validator_a"])\n\n[clazz().validate() for clazz in validator_classes]\n# -> validateB!!\n# -> validateC!!\n\n# Pattern2: \'load_config\'\nfrom autoload import load_config\n\n@load_config(load=False)\nclass ValidatorA:\n  def validate(self):\n    print("validateA!!")\n\nvalidator_classes = loader.load_classes("main")\n\n[clazz().validate() for clazz in validator_classes]\n# -> validateB!!\n# -> validateC!!\n```\nThis function will check directory structure recursively if you specify `recursive=True`. \n```python\n# \'recursive=False\' is default.\n# In this case, the loader will also check \'pkg/main/sub/\'.\nvalidator_classes = loader.load_classes("main", recursive=True)\n\n[clazz().validate() for clazz in validator_classes]\n# -> validateA!!\n# -> validateB!!\n# -> validateC!!\n# -> validateD!!\n# -> validateE!!\n```\nYou can specify `pkg_name` as below.\n```python\nloader.load_classes("main/validator_a.py")\nloader.load_classes("main.validator_a")\nloader.load_classes("./main/validator_a")\nloader.load_classes(".main.validator_a")\nloader.load_classes("main.sub.validator_d")\nloader.load_classes("./main/sub/validator_d")\nloader.load_classes("../otherpkg")\nloader.load_classes("..otherpkg")\n```\n\n#### load_functions\n```\nload_functions(\n    src: str,\n    excludes: Iterable[str] = (),\n    recursive: bool = False,\n) -> Tuple[Callable]:\n```\nThis method read the Python package or module and return the tuple of functions.\nThe usage is the same as `load_classes`.\n\n##### NOTE\n- To search class or function, **You must match the name of file, and the one of class or function.**\nFor example, if you named the file `test_module.py`, you must name the class `TestModule` or the function `test_module`.\nWhen you want to customize their name, use `@load_config` decorator.\n    - validator_a.py\n    ```python\n    from autoload import load_config\n  \n    @load_config()\n    class CustomValidator:\n        def validate(self):\n            print("validateA!!")\n    ```\n- You can also control the order of loaded class objects using `@load_config` decorator.\n    - validator_a.py\n    ```python\n    from autoload import load_config\n  \n    # sort in ascending order\n    @load_config(order=1)\n    class ValidatorA:\n        def validate(self):\n            print("validateA!!")\n    ```\n- If you decorate some classes or functions with `@load_config`, the loader will load them.\n  However, initialized `strict=True`, the loader denies multiple loading as below.\n  - pkg/validator_a.py\n  ```python\n  from autoload import load_config\n\n  # This will be loaded because of name\'s rule.\n  class ValidatorA:\n    def validate(self):\n        print("validateA!!")\n  \n  # Anything goes.\n  @load_config(order=2)\n  class Foo:\n    pass\n  ```\n  \n  - main.py\n  ```python\n  from autoload import ModuleLoader\n  from autoload.exception import LoaderStrictModeError\n  \n  loader = ModuleLoader()\n  # return ValidatorA and Foo class objects.\n  classes = loader.load_classes("pkg")\n\n  # ModuleLoader strictly try to load a class or function object\n  # per a Python module on a basis of its name.\n  strict_loader = ModuleLoader(strict=True)\n  try:\n    classes = strict_loader.load_classes("pkg")\n  except LoaderStrictModeError as e:\n    print(e)\n  # -> Loader can only load a \'ValidatorA\' class in validator_a module.\n  # -> Please check \'Foo\' in validator_a module.\n  ```\n#### load_class\n```\nload_class(file_name: str)\n```\nThis method read the Python file and return the class object.\n- Directory\n```\nproject/\n  ├ example.py\n  └ validator.py\n```\n- validator.py\n```python\nclass Validator:\n    def validate(self):\n        print("validate!!")\n```\n- example.py\n```python\nloader = ModuleLoader()\nclazz = loader.load_class("validator")\nclazz().validate()\n# -> validate!!\n```\nHow to specify `file_name` is the same as that of `load_classes`.\n\n#### load_function\n```\nload_function(file_name: str)\n```\nThis method read the Python file and return a function object.\nThe usage is the same as `load_class`.\n\n## License\nReleased under the MIT license.',
    'author': 'Hiroki Miyaji',
    'author_email': 'nukoprogramming@gmail.com',
    'maintainer': 'Hiroki Miyaji',
    'maintainer_email': 'nukoprogramming@gmail.com',
    'url': 'https://github.com/hiroki0525/autoload_module',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



__all__ = "load_config"

from typing import Callable, Optional

from autoload._globals import Class_Or_Func


def load_config(
    order: Optional[int] = None, load: bool = True
) -> Callable[[Class_Or_Func], Class_Or_Func]:
    def decorator(resource: Class_Or_Func):
        if order:
            setattr(resource, "_load_order", order)
        setattr(resource, "_load_flg", load)
        return resource

    return decorator
